"""Router dependency injections."""

# pylint: disable=R0903

from typing import Annotated

import requests
from fastapi import Depends
from openbb_core.provider.utils.helpers import get_requests_session

Session = Annotated[requests.Session, Depends(get_requests_session)]



"""{{ cookiecutter.package_name }} OpenBB Platform Provider."""

from openbb_core.provider.abstract.provider import Provider
from {{cookiecutter.package_name}}.providers.{{cookiecutter.provider_name}}.models.example import ExampleFetcher
from {{cookiecutter.package_name}}.providers.{{cookiecutter.provider_name}}.models.ohlc_example import {{cookiecutter.provider_name.replace('_', ' ').title().replace(' ', '')}}EquityHistoricalFetcher



{{cookiecutter.provider_name}}_provider = Provider(
    name="{{cookiecutter.provider_name}}",
    description="Data provider for {{cookiecutter.project_name}}.",
    # Only add 'credentials' if they are needed.
    # For multiple login details, list them all here.
    # credentials=["api_key"],
    website="https://{{cookiecutter.project_tag}}.com",
    # Here, we list out the fetchers showing what our provider can get.
    # The dictionary key is the fetcher's name, used in the `../routers/router.py`.
    fetcher_dict={
        "EquityHistorical": {{cookiecutter.provider_name.replace('_', ' ').title().replace(' ', '')}}EquityHistoricalFetcher,
        "Example": ExampleFetcher,
    }
)



"""{{cookiecutter.router_name}} router command example."""

# pylint: disable=unused-argument

from openbb_core.app.model.command_context import CommandContext
from openbb_core.app.model.example import APIEx, PythonEx
from openbb_core.app.model.obbject import OBBject
from openbb_core.app.provider_interface import ExtraParams, ProviderChoices, StandardParams
from openbb_core.app.query import Query
from openbb_core.app.router import Router
from pydantic import BaseModel

# Example dependency injection yielding a configured requests.Session object.
from {{cookiecutter.package_name}}.routers.depends import Session

# The prefix extension's prefix is determined by the `pyproject.toml` EntryPoint assignment.
# Assign a prefix only if this is a sub-router.
router = Router(prefix="")


@router.command(
    methods=["GET"],
    examples=[
        PythonEx(
            description="Here is an example for using this endpoint.",
            code=[
                "obb.{{ cookiecutter.router_name }}.get_example(symbol='AAPL')",
            ]
        )
    ]
)
async def get_example(session: Session, symbol: str = "AAPL") -> OBBject[dict]:
    """Get options data."""
    url = f"https://www.cboe.com/education/tools/trade-optimizer/symbol-info?symbol={symbol}"
    response = session.get(url)
    response.raise_for_status()
    data = response.json()

    return OBBject(results=data["details"])


@router.command(methods=["POST"])
async def post_example(
    data: BaseModel,  # These are body parameters.
    flag: bool = False,  # These are query parameters.
) -> OBBject[dict]:
    """Calculate mid and spread."""

    bid = getattr(data, "bid_col", 0)
    ask = getattr(data, "ask_col", 0)
    mid = (bid + ask) / 2
    spread = ask - bid

    return OBBject(results={"mid": mid, "spread": spread, "flag": flag})


@router.command(model="Example")
async def model_example(
    cc: CommandContext,
    provider_choices: ProviderChoices,
    standard_params: StandardParams,
    extra_params: ExtraParams,
) -> OBBject[BaseModel]:
    """Example Data."""
    return await OBBject.from_query(Query(**locals()))


# If you had another provider installed that mapped to this model - i.e, `openbb-fmp`
# they will be added to this endpoint.
@router.command(model="EquityHistorical")
async def candles(
    cc: CommandContext,
    provider_choices: ProviderChoices,
    standard_params: StandardParams,
    extra_params: ExtraParams,
) -> OBBject: # results type is inferred from Fetcher annotations.
    """Example Data."""
    return await OBBject.from_query(Query(**locals()))



"""Root configuration for pytest."""

# flake8: noqa: S101
# pylint: disable=unused-argument,unused-import

import os
from pathlib import Path

import pytest  # noqa: F401

ROOT_DIR = Path(__file__).parent


def pytest_configure():
    """Set environment variables for testing."""
    os.environ["OPENBB_AUTO_BUILD"] = "true"


def pytest_collection_modifyitems(config, items):
    """Modify test collection to ensure cleanup-dependent tests run first."""
    # Find tests that should run early (checking clean state)
    early_tests: list = []
    other_tests: list = []

    for item in items:
        # Tests that check repository state should run first
        if item.get_closest_marker("order"):
            early_tests.append(item)
        else:
            other_tests.append(item)

    # Sort early tests by their order marker if present
    early_tests.sort(
        key=lambda x: (
            getattr(x.get_closest_marker("order"), "args", [999])[0]
            if x.get_closest_marker("order")
            else 999
        )
    )

    # Reorder: early tests first, then others
    items[:] = early_tests + other_tests



"""Example Data Integration With Standard Model.

This file shows an example of how to integrate this provider with ends available to other providers.
"""

# pylint: disable=unused-argument
from typing import Any

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.equity_historical import (
    EquityHistoricalData,
    EquityHistoricalQueryParams,
)
from pydantic import Field, field_validator


class {{cookiecutter.provider_name.replace('_', ' ').capitalize().replace(' ', '')}}EquityHistoricalQueryParams(EquityHistoricalQueryParams):
    """Example provider query.

    The standard model here comes with parameters for symbol, start_date, and end_date.
    """

    custom_param: str | None = Field(
        default=None, description="Some optional parameter"
    )


class {{cookiecutter.provider_name.replace('_', ' ').capitalize().replace(' ', '')}}EquityHistoricalData(EquityHistoricalData):
    """Sample provider data.

    The standard model has these fields,
    so we use __alias_dict__ to map them.
    We only need to add fields not in the inherited model, or to override.
    """

    __alias_dict__ = {
        "date": "d",
        "open": "o",
        "high": "h",
        "low": "l",
        "close": "c",
        "volume": "v",
        "custom_field": "f",
    }
    custom_field: str | None = Field(default=None, description="Some optional field")

    @field_validator("custom_field", mode="before", check_fields=False)
    @classmethod
    def _validate_custom_field(cls, v):
        """Validate the custom field."""
        return v if v else "Data validator replaced None."


class {{cookiecutter.provider_name.replace('_', ' ').capitalize().replace(' ', '')}}EquityHistoricalFetcher(
    Fetcher[
        {{cookiecutter.provider_name.replace('_', ' ').capitalize().replace(' ', '')}}EquityHistoricalQueryParams,
        list[{{cookiecutter.provider_name.replace('_', ' ').capitalize().replace(' ', '')}}EquityHistoricalData],
    ]
):
    """Example Fetcher class.

    This class is responsible for the actual data retrieval.
    """

    @staticmethod
    def transform_query(params: dict[str, Any]) -> {{cookiecutter.provider_name.replace('_', ' ').capitalize().replace(' ', '')}}EquityHistoricalQueryParams:
        """Define example transform_query.

        Here we can pre-process the query parameters and add any extra parameters that
        will be used inside the extract_data method.
        """
        return {{cookiecutter.provider_name.replace('_', ' ').capitalize().replace(' ', '')}}EquityHistoricalQueryParams(**params)

    # Note the use of async here. Make the Fetcher async with this small change.
    @staticmethod
    async def aextract_data(
        query: {{cookiecutter.provider_name.replace('_', ' ').capitalize().replace(' ', '')}}EquityHistoricalQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list[dict]:
        """Define example extract_data.

        Here we make the actual request to the data provider and receive the raw data.
        If you said your Provider class needs credentials you can get them here.
        """
        api_key = (
            credentials.get("{{ cookiecutter.provider_name }}_api_key") if credentials else ""
        )

        # Here we mock an example_response for brevity.
        # Show model validation by only returning one row of custom_field
        # Show model validation by only returning one row of custom_field
        example_response = [
            {
                "o": 2,
                "h": 5,
                "l": 1,
                "c": 4,
                "v": 5,
                "d": "August 23, 2023",
                "f": query.custom_param,
            },
            {
                "o": 4,
                "h": 7,
                "l": 3,
                "c": 6,
                "v": 10,
                "d": "August 24, 2023",
                "f": None,
            },
        ]

        return example_response

    @staticmethod
    def transform_data(
        query: {{cookiecutter.provider_name.replace('_', ' ').capitalize().replace(' ', '')}}EquityHistoricalQueryParams,
        data: list[dict],
        **kwargs: Any
    ) -> list[{{cookiecutter.provider_name.replace('_', ' ').capitalize().replace(' ', '')}}EquityHistoricalData]:
        """Define example transform_data.

        Right now, we're converting the data to fit our desired format.
        You can apply other transformations to it here.
        """
        return [{{cookiecutter.provider_name.replace('_', ' ').capitalize().replace(' ', '')}}EquityHistoricalData.model_validate(d) for d in data]



"""OpenBB Cookiecutter Template."""

from pathlib import Path

__version__ = "0.4.0"


def get_template_path() -> Path:
    """Return the path to the cookiecutter template directory."""
    return Path(__file__).parent / "template"



"""OpenBB Platform Extension post-generation script."""

import re
import sys

MODULE_REGEX = r"^[_a-zA-Z][_a-zA-Z0-9]+$"

MODULE_NAME = "{{ cookiecutter.package_name }}"
PROVIDER_NAME = "{{ cookiecutter.provider_name }}" or ""
ROUTER_NAME = "{{ cookiecutter.router_name }}" or ""
OBBJECT_NAME = "{{ cookiecutter.obbject_name }}" or ""

if not re.match(MODULE_REGEX, MODULE_NAME):
    print(f"ERROR: {MODULE_NAME} is not a valid Python package name.")

    sys.exit(1)

if PROVIDER_NAME and not re.match(MODULE_REGEX, PROVIDER_NAME):
    print(f"ERROR: {PROVIDER_NAME} should be in lower snakecase.")

    sys.exit(1)

if ROUTER_NAME and not re.match(MODULE_REGEX, ROUTER_NAME):
    print(f"ERROR: {ROUTER_NAME} should be in lower snakecase.")

    sys.exit(1)

if OBBJECT_NAME and not re.match(MODULE_REGEX, OBBJECT_NAME):
    print(f"ERROR: {OBBJECT_NAME} should be in lower snakecase.")

    sys.exit(1)



"""{{ cookiecutter.package_name }} OBBject Extension - {{ cookiecutter.obbject_name }}"""

# pylint: disable=W0613,R0903

import threading
import time

from openbb_core.app.model.extension import Extension
from openbb_core.app.model.obbject import OBBject

# Extensions are registered as OBBject accessors.
# It can be a class, or it can be a callable method.
ext = Extension(
    name="to_string",
    description="An OBBject extension that converts the results to a string representation.",
)

# If it is a function, no parameters will be accepted.
# The function will execute like a property method.
# The accessor is called when the namespace is entered.
@ext.obbject_accessor
def to_string(obbject, **kwargs) -> str:
    """OBBject accessor providing a "to_string" method."""
    return obbject.model_dump_json(exclude_none=True, exclude_unset=True, include="results")

# We ignore this OpenBBWarning: Skipping '{{ cookiecutter.obbject_name }}', name already in user.

class_ext = Extension(
    name="{{ cookiecutter.obbject_name }}",
    description="An OBBject extension with namespace."
)

@class_ext.obbject_accessor
class OBBjectExtension:
    """OBBject Extension Template."""

    def __init__(self, obbject: OBBject):
        """Initialize the extension."""
        self._obbject = obbject

    def hello_world(self, **kwargs):
        """Say hello from the OBBject extension."""
        print(f"Hello from the OBBject instance! \n\n{repr(self._obbject)}")  # noqa

## Non-blocking OBBject Extension Example
## Uncomment to use
#nonblocking_plugin = Extension(
#    name="nonblocking_plugin",
#    description="An on-command-output plugin simulating an extensive task performed in a separate thread.",
#    on_command_output=True,  # Must be set as True
#    command_output_paths=["/{{cookiecutter.router_name}}/candles"],
#    immutable=True,  # Set to `True` for parallel processing.
#    results_only=False,  # Use this as a flag to return only the "results" portion of the OBBject.
#)


#def _expensive_operation_worker(serialized_obbject: dict):
#    """Simulate a long-running task without blocking the caller."""
#    working_copy = OBBject(**serialized_obbject)
#    print("\nThis is the deserialized OBBject in the non-blocking thread.")
#    print(working_copy.__repr__())
#    for i in range(10):
#        print(str(i) + " seconds remaining...")
#        time.sleep(1)
#    print("Expensive operation is now complete.")


#@nonblocking_plugin.obbject_accessor
#def empty_plugin_function(obbject):  # This can also be an async function.
#    """Simulated on_commnd_output function that executes an expensive task
#    in a non-blocking thread."""
#    print(
#        "Serializing the obbject and passing to a new thread.\n"
#        f"Command executed: {obbject.extra['metadata']}\n"
#    )
#    print(
#        "Simulating an expensive task that is non-blocking and allows the function to return."
#    )
#    threading.Thread(
#        target=_expensive_operation_worker,
#        args=(obbject.model_dump(),),
#        name="empty-plugin-expensive-operation",
#        daemon=False,
#    ).start()



"""Views for the {{ cookiecutter.router_name }} Extension."""

# flake8: noqa: PLR0912
# pylint: disable=import-outside-toplevel,too-few-public-methods

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from openbb_charting.core.openbb_figure import OpenBBFigure  # `openbb-charting` is Optional for the user.


# You can make charts that are returned when you map a function
# route in lower_snake_case to this class using static methods.
# You can return whatever format you like, but it must be JSON-serializable.
# The returned tuple object gets added to the response object from the application.
# The charting extension itself is accessible under the `charting` namespace.
# While the finished chart is under the `chart` object of the OBBject response output.
# The application will check if the user has `openbb-charting` installed on run.
# If not, the views are not added to the application.


class {{cookiecutter.router_name.replace("_", " ").title().replace(" ", "")}}Views:
    """{{ cookiecutter.router_name }} Views."""

    @staticmethod
    def {{cookiecutter.router_name}}_candles(**kwargs) -> tuple["OpenBBFigure", dict[str, Any]]:
        """Create a chart that will return to the API as a JSON-encoded string."""
        # Keep imports here so they are imported only at function run.
        from openbb_charting.core.openbb_figure import OpenBBFigure

        data = kwargs["obbject_item"]  # This is where the data will always be.

        print(data)

        fig = OpenBBFigure()

        fig.add_bar(x=[d.date for d in data], y=[d.high for d in data])
        content = fig.show(external=True).to_plotly_json()
        # fig should be the binary Python object of the chart
        # content should be a JSON-serialized version ready for the frontend to render.
        return fig, content



"""CLI for OpenBB Cookiecutter template."""

# pylint: disable=W0718

import argparse
import sys

from cookiecutter.main import cookiecutter

from . import get_template_path


def main(argv: list | None = None) -> int:
    """Run the OpenBB cookiecutter template.

    Args:
        argv: Command line arguments (defaults to sys.argv[1:])

    Returns:
        Exit code (0 for success, non-zero for error)
    """
    parser = argparse.ArgumentParser(
        description="Generate an OpenBB Platform extension from template"
    )
    parser.add_argument(
        "-o",
        "--output-dir",
        default=".",
        help="Where to output the generated project (default: current directory)",
    )
    parser.add_argument(
        "--no-input",
        action="store_true",
        help="Do not prompt for parameters and use defaults",
    )
    parser.add_argument(
        "-f", "--overwrite-if-exists", action="store_true", help="Overwrite if exists"
    )
    parser.add_argument(
        "--extra-context",
        action="append",
        metavar="KEY=VALUE",
        help="Extra context variables (can be used multiple times)",
    )

    args = parser.parse_args(argv)

    # Build extra context from arguments
    extra_context = {}
    if args.extra_context:
        for item in args.extra_context:
            if "=" not in item:
                print(f"Error: extra-context must be in KEY=VALUE format: {item}")
                return 1
            key, value = item.split("=", 1)
            extra_context[key] = value

    # Get the bundled template path
    template_path = get_template_path()

    try:
        cookiecutter(
            str(template_path),
            output_dir=args.output_dir,
            no_input=args.no_input,
            overwrite_if_exists=args.overwrite_if_exists,
            extra_context=extra_context if extra_context else None,
        )
        return 0
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)  # noqa
        return 1


if __name__ == "__main__":
    sys.exit(main())



"""Example Data Integration.

The OpenBB Platform gives developers easy tools for integration.

To use it, developers should:
1. Define the request/query parameters.
2. Define the resulting data schema.
3. Define how to fetch raw data.

First 2 steps make sure developers really get to know their data.
This is called the "Know Your Data" principle.

Note: The format of the QueryParams and Data is defined by a pydantic model that can
be entirely custom, or inherit from the OpenBB standardized models.

This file shows an example of how to integrate data from a provider.
"""
# pylint: disable=unused-argument
from typing import Any, Optional

from openbb_core.provider.abstract.data import Data
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.abstract.query_params import QueryParams
from pydantic import Field


class ExampleQueryParams(QueryParams):
    """Example provider query.

    This is the definition of our query parameters that are specific to this provider.
    We use this class to create our own parameters that will provided as input to the
    command.
    """

    symbol: str = Field(description="Symbol to query.")


class ExampleData(Data):
    """Sample provider data.

    The fields are displayed as-is in the output of the command. In this case, its the
    Open, High, Low, Close and Volume data.
    """

    o: float = Field(description="Open price.")
    h: float = Field(description="High price.")
    l: float = Field(description="Low price.")
    c: float = Field(description="Close price.")
    v: float = Field(description="Volume.")
    d: str = Field(description="Date")


class ExampleFetcher(
    Fetcher[
        ExampleQueryParams,
        list[ExampleData],
    ]
):
    """Example Fetcher class.

    This class is responsible for the actual data retrieval.
    """

    @staticmethod
    def transform_query(params: dict[str, Any]) -> ExampleQueryParams:
        """Define example transform_query.

        Here we can pre-process the query parameters and add any extra parameters that
        will be used inside the extract_data method.
        """
        return ExampleQueryParams(**params)

    @staticmethod
    def extract_data(
        query: ExampleQueryParams,
        credentials: dict[str, str] | None,
        **kwargs: Any,
    ) -> list[dict]:
        """Define example extract_data.

        Here we make the actual request to the data provider and receive the raw data.
        If you said your Provider class needs credentials you can get them here.
        """
        api_key = (
            credentials.get("{{cookiecutter.package_name}}_api_key")
            if credentials
            else ""
        )

        # Here we mock an example_response for brevity.
        example_response = [
            {
                "o": 2,
                "h": 5,
                "l": 1,
                "c": 4,
                "v": 5,
                "d": "August 23, 2023",
            },
            {
                "o": 4,
                "h": 7,
                "l": 3,
                "c": 6,
                "v": 10,
                "d": "August 24, 2023",
            },
        ]

        return example_response

    @staticmethod
    def transform_data(
        query: ExampleQueryParams, data: list[dict], **kwargs: Any
    ) -> list[ExampleData]:
        """Define example transform_data.

        Right now, we're converting the data to fit our desired format.
        You can apply other transformations to it here.
        """
        return [ExampleData(**d) for d in data]
